' ============
' Voxel Editor
' ============
' Written by Richard Averill, 1999.
' E-mail me at nigelolympus@tinet.ie

' --------------------------
' Declare SUBs and FUNCTIONs
' --------------------------

DECLARE SUB MouseDriver (ax%, bx%, cx%, dx%)
DECLARE SUB Drawegrid ()
DECLARE SUB Drawiview ()
DECLARE SUB ButtonWait ()
DECLARE SUB Drawgridbuts ()

DEFINT A-Z

REM $DYNAMIC

CHDIR "C:\RICHARD\QBASIC45\PROGRAMS\"

' -----------------
' Declare constants
' -----------------

CONST radian = (3.141593# / 180)

' -------------------
' Dimension variables
' -------------------

DIM SHARED voxel%(16, 16, 16)
DIM SHARED mouse$: mouse$ = SPACE$(57)
DIM SHARED col%, gridx%, gridy%, gridz%, perang%, grid%
DIM perbuff%(5000)
DIM pal%(512)

' ------------
' Screen setup
' ------------

CLS : SCREEN 13

' -----------------------
' Read machine code array
' -----------------------

FOR i% = 1 TO 57
  
   READ a$
   H$ = CHR$(VAL("&H" + a$))
   MID$(mouse$, i%, 1) = H$

NEXT i%

' -----------------
' Machine code data
' -----------------

DATA 55,89,E5,8B,5E,0C,8B,07,50,8B,5E,0A,8B,07,50,8B
DATA 5E,08,8B,0F,8B,5E,06,8B,17,5B,58,1E,07,CD,33,53
DATA 8B,5E,0C,89,07,58,8B,5E,0A,89,07,8B,5E,08,89,0F
DATA 8B,5E,06,89,17,5D,CA,08,00

RESTORE

' --------------------
' Initialize variables
' --------------------

grid% = 0: col% = 0: gridx% = 0: gridy% = 0
gridz% = 15: perang% = 0

' ------------
' Draw palette
' ------------

FOR i% = 0 TO 31
   FOR j% = 0 TO 7

      x1% = (272 + (j% * 6)): x2% = (x1% + 5)
      y1% = (i% * 5): y2% = (y1% + 4)

      LINE (x1%, y1%)-(x2%, y2%), (i% * 8) + j%, BF

   NEXT j%
NEXT i%

PSET (274, 2), 15

' -----------------
' Draw editing grid
' -----------------

FOR i% = 0 TO 16

   LINE (174, 102 + (i% * 6))-(270, 102 + (i% * 6)), 15
   LINE (174 + (i% * 6), 102)-(174 + (i% * 6), 198), 15

NEXT i%

Drawegrid

' -------------------------------
' Draw frame for perspective view
' -------------------------------

LINE (174, 0)-(270, 100), 15, B

' ------------------------------
' Draw frame for selected colour
' ------------------------------

LINE (272, 160)-(319, 198), 15, B

' --------------------------
' Draw grid movement buttons
' --------------------------

LINE (150, 121)-(169, 140), 8, BF
LINE (149, 120)-(170, 141), 15, B

LINE (151, 130)-(168, 130), 15
LINE (159, 122)-(159, 139), 15

LINE (150, 145)-(169, 164), 8, BF
LINE (149, 144)-(170, 165), 15, B
LINE (151, 154)-(168, 154), 15

LOCATE 23, 18: PRINT gridx%

' ---------------------------
' Draw grid selection buttons
' ---------------------------

Drawgridbuts

' -------------------
' Draw isometric view
' -------------------

Drawiview

' -----------------------------
' Draw load, save and new boxes
' -----------------------------

LOCATE 18, 2: PRINT "Load"
LOCATE 20, 2: PRINT "Save"
LOCATE 18, 8: PRINT "New"
LOCATE 20, 8: PRINT "Quit"

LINE (4, 134)-(42, 144), 15, B
LINE (4, 150)-(42, 160), 15, B
LINE (52, 134)-(92, 144), 15, B
LINE (52, 150)-(92, 160), 15, B

' -----------
' Print title
' -----------

LOCATE 2, 2: COLOR 15: PRINT "Q-Basic Voxel Editor"
LINE (5, 5)-(170, 18), 15, B

' --------------------
' Initialize the mouse
' --------------------

ax% = 0
MouseDriver ax%, 0, 0, 0
MouseInit% = ax%

IF NOT MouseInit% THEN
   PRINT "Mouse not found"
   END
END IF

' ----------------------
' Make the mouse visible
' ----------------------

ax% = 1
MouseDriver ax%, 0, 0, 0

' ------------------------
' Set the mouse boundaries
' ------------------------

ax% = 7: cx% = 0: dx% = (319 * 2)
MouseDriver ax%, 0, cx%, dx%

ax% = 8: cx% = 0: dx% = 199
MouseDriver ax%, 0, cx%, dx%

' ------------
' Main routine
' ------------

DO
  
   ' ---------------------
   ' Read the mouse status
   ' ---------------------

   ax% = 3
   MouseDriver ax%, bx%, cx%, dx%
 
   lb% = ((bx% AND 1) <> 0)
   Rb% = ((bx% AND 2) <> 0)
   xmouse% = (cx% / 2)
   ymouse% = dx%

   ' -----------------
   ' Check left button
   ' -----------------

   SELECT CASE lb%
     
      CASE -1
        
         ' ------------------------------
         ' Check if palette is clicked on
         ' ------------------------------

         IF (xmouse% > 271) AND (ymouse% < 160) THEN
            colx% = INT((xmouse% - 272) / 6)
            coly% = INT(ymouse% / 5)
            col% = (coly% * 8) + colx%
           
            ' ------------------------
            ' Make the mouse invisible
            ' ------------------------

            ax% = 2
            MouseDriver ax%, 0, 0, 0

            ' --------------------------------
            ' Fill in the colour selection box
            ' --------------------------------

            LINE (273, 161)-(318, 197), col%, BF

            ' ----------------------
            ' Make the mouse visible
            ' ----------------------

            ax% = 1
            MouseDriver ax%, 0, 0, 0
           
         END IF

         ' -----------------------------------
         ' Check if editing grid is clicked on
         ' -----------------------------------

         IF (xmouse% > 174) AND (xmouse% < 270) THEN
            IF (ymouse% > 102) AND (ymouse% < 198) THEN
               IF ((xmouse% - 174) MOD 6 <> 0) AND ((ymouse% - 102) MOD 6 <> 0) THEN

                  ' ---------------------
                  ' Work out co-ordinates
                  ' ---------------------

                  x% = INT((xmouse% - 174) / 6)
                  y% = INT((ymouse% - 102) / 6)

                  x1% = (175 + (x% * 6))
                  y1% = (103 + (y% * 6))

                  ' ------------------------
                  ' Make the mouse invisible
                  ' ------------------------

                  ax% = 2
                  MouseDriver ax%, 0, 0, 0

                  ' -------------------------------------
                  ' Fill in the block in the editing grid
                  ' -------------------------------------

                  LINE (x1%, y1%)-(x1% + 4, y1% + 4), col%, BF
                  IF (col% = 0) THEN PSET (x1% + 2, y1% + 2), 15

                  ' ----------------------
                  ' Make the mouse visible
                  ' ----------------------

                  ax% = 1
                  MouseDriver ax%, 0, 0, 0

                  ' ---------------------
                  ' Alter the voxel array
                  ' ---------------------

                  SELECT CASE grid%
                     CASE 0: voxel%(gridx%, x%, (15 - y%)) = col%
                     CASE 1: voxel%(x%, (15 - gridy%), (15 - y%)) = col%
                     CASE 2: voxel%((15 - y%), x%, gridz%) = col%
                  END SELECT

                  ' ---------------------
                  ' Redraw isometric view
                  ' ---------------------

                  Drawiview

               END IF
            END IF
         END IF

         ' -------------------------------------------
         ' Check if grid movement boxes are clicked on
         ' -------------------------------------------

         IF (xmouse% > 148) AND (xmouse% < 171) THEN
            IF (ymouse% > 119) AND (ymouse% < 142) THEN

               ' -----------------
               ' Move up one space
               ' -----------------

               SELECT CASE grid%
                  CASE 0
                     gridx% = (gridx% + 1)
                     IF (gridx% > 15) THEN gridx% = 0
                     LOCATE 23, 18: PRINT gridx%
                  CASE 1
                     gridy% = (gridy% + 1)
                     IF (gridy% > 15) THEN gridy% = 0
                     LOCATE 23, 18: PRINT gridy%
                  CASE 2
                     gridz% = (gridz% + 1)
                     IF (gridz% > 15) THEN gridz% = 0
                     LOCATE 23, 18: PRINT gridz%
               END SELECT

               Drawegrid
               Drawiview
               ButtonWait

            ELSEIF (ymouse% > 143) AND (ymouse% < 166) THEN

               ' -------------------
               ' Move down one space
               ' -------------------

               SELECT CASE grid%
                  CASE 0
                     gridx% = (gridx% - 1)
                     IF (gridx% < 0) THEN gridx% = 15
                     LOCATE 23, 18: PRINT gridx%
                  CASE 1
                     gridy% = (gridy% - 1)
                     IF (gridy% < 0) THEN gridy% = 15
                     LOCATE 23, 18: PRINT gridy%
                  CASE 2
                     gridz% = (gridz% - 1)
                     IF (gridz% < 0) THEN gridz% = 15
                     LOCATE 23, 18: PRINT gridz%
               END SELECT

               Drawegrid
               Drawiview
               ButtonWait

           END IF
        END IF

        ' ----------------------------------------------
        ' Check if grid selection buttons are clicked on
        ' ----------------------------------------------

        IF (xmouse% > 114) AND (xmouse% < 132) THEN
           IF (ymouse% > 133) AND (ymouse% < 145) THEN
              grid% = 0
              Drawgridbuts
              Drawegrid
           ELSEIF (ymouse% > 149) AND (ymouse% < 161) THEN
              grid% = 1
              Drawgridbuts
              Drawegrid
           ELSEIF (ymouse% > 165) AND (ymouse% < 177) THEN
              grid% = 2
              Drawgridbuts
              Drawegrid
           END IF
        END IF

        ' --------------------------------------------
        ' Check if file operations buttons are clicked
        ' --------------------------------------------

        IF (xmouse% > 3) AND (xmouse% < 43) THEN
           IF (ymouse% > 133) AND (ymouse% < 145) THEN
             
              ' ----
              ' Load
              ' ----

              LOCATE 23, 2: INPUT ">", file$
             
              OPEN file$ FOR BINARY AS #1
             
              a% = 0

              FOR x% = 0 TO 15
                 FOR y% = 0 TO 15
                    FOR z% = 0 TO 15

                       a% = (a% + 2)

                       GET #1, a%, voxel%(x%, y%, z%)

                    NEXT z%
                 NEXT y%
              NEXT x%

              CLOSE

              Drawegrid
              Drawiview

              LOCATE 23, 2: PRINT SPACE$(13)

           ELSEIF (ymouse% > 149) AND (ymouse% < 161) THEN
              
              ' ----
              ' Save
              ' ----

              LOCATE 23, 2: INPUT ">", file$
              OPEN file$ FOR BINARY AS #1
            
              a% = 0

              FOR x% = 0 TO 15
                 FOR y% = 0 TO 15
                    FOR z% = 0 TO 15

                       a% = (a% + 2)
                       PUT #1, a%, voxel%(x%, y%, z%)

                    NEXT z%
                 NEXT y%
              NEXT x%

               CLOSE

               LOCATE 23, 2: PRINT SPACE$(13)

            END IF
         END IF
     
         ' --------------------------
         ' Check new and quit buttons
         ' --------------------------

         IF (xmouse% > 51) AND (xmouse% < 93) THEN
            IF (ymouse% > 133) AND (ymouse% < 145) THEN

               LOCATE 23, 2: INPUT "New?", sure$
           
               IF LEFT$(sure$, 1) = "y" OR LEFT$(sure$, 1) = "Y" THEN
                  ERASE voxel%
                  REDIM SHARED voxel%(16, 16, 16)
                  Drawegrid
                  Drawiview
               END IF

               LOCATE 23, 2: PRINT SPACE$(13)

            ELSEIF (ymouse% > 149) AND (ymouse% < 161) THEN
               
               LOCATE 23, 2: INPUT "Quit?", sure$
          
               IF LEFT$(sure$, 1) = "y" OR LEFT$(sure$, 1) = "Y" THEN
                  CLS : END
               END IF
              
               LOCATE 23, 2: PRINT SPACE$(13)

            END IF
         END IF

      CASE ELSE

   END SELECT

   ' ----------------------
   ' Check the right button
   ' ----------------------

   SELECT CASE Rb%
     
      CASE -1

         ' -----------------------------------
         ' Check if editing grid is clicked on
         ' -----------------------------------

         IF (xmouse% > 174) AND (xmouse% < 270) THEN
            IF (ymouse% > 102) AND (ymouse% < 198) THEN
               IF ((xmouse% - 174) MOD 6 <> 0) AND ((ymouse% - 102) MOD 6 <> 0) THEN

                  ' ---------------------
                  ' Work out co-ordinates
                  ' ---------------------

                  x% = INT((xmouse% - 174) / 6)
                  y% = INT((ymouse% - 102) / 6)

                  SELECT CASE grid%
                     CASE 0: col% = voxel%(gridx%, x%, (15 - y%))
                     CASE 1: col% = voxel%(x%, (15 - gridy%), (15 - y%))
                     CASE 2: col% = voxel%((15 - y%), x%, gridz%)
                  END SELECT

                  ' ------------------------
                  ' Make the mouse invisible
                  ' ------------------------

                  ax% = 2
                  MouseDriver ax%, 0, 0, 0

                  ' --------------------------------
                  ' Fill in the colour selection box
                  ' --------------------------------

                  LINE (273, 161)-(318, 197), col%, BF

                  ' ----------------------
                  ' Make the mouse visible
                  ' ----------------------

                  ax% = 1
                  MouseDriver ax%, 0, 0, 0

               END IF
            END IF
         END IF

      CASE ELSE
  
   END SELECT

   ' -------------------------
   ' Draw the perspective view
   ' -------------------------

   ERASE perbuff%
   REDIM perbuff%(5000)
   perbuff%(0) = (95 * 8)
   perbuff%(1) = 99

   DEF SEG = VARSEG(perbuff%(2))
   peroffs& = VARPTR(perbuff%(2))

   perrad# = (perang% * radian)

   FOR y% = 0 TO 15
      FOR x% = 0 TO 15

         SELECT CASE perang%
            CASE 0 TO 89
               x2% = x%
               y2% = y%
            CASE 90 TO 179
               x2% = x%
               y2% = (15 - y%)
            CASE 180 TO 269
               x2% = (15 - x%)
               y2% = (15 - y%)
            CASE 270 TO 359
               x2% = (15 - x%)
               y2% = y%
         END SELECT

         x3% = ((x2% - 7) * 3)
         y3% = ((y2% - 7) * 3)

         x4! = (x3% * COS(perrad#)) - (y3% * SIN(perrad#))
         y4! = (x3% * SIN(perrad#)) + (y3% * COS(perrad#))
    
         x5% = INT(x4! + 48)
         y5% = INT(y4! * .7 + 72)
                                  
         offset& = INT((y5% * 95) + x5%)

         FOR z% = 0 TO 15
       
            k% = voxel%(x2%, y2%, z%)
           
            IF (k% = 56) THEN
               k% = 56
               POKE (offset& - 96), k% + (PEEK(offset& - 96) AND 7)
               POKE (offset& - 95), k% + (PEEK(offset& - 95) AND 7)
               POKE (offset& - 94), k% + (PEEK(offset& - 94) AND 7)
               POKE (offset& - 1), k% + (PEEK(offset& - 1) AND 7)
               POKE offset&, k% + (PEEK(offset&) AND 7)
               POKE (offset& + 1), k% + (PEEK(offset& + 1) AND 7)
               POKE (offset& + 94), k% + (PEEK(offset& + 94) AND 7)
               POKE (offset& + 95), k% + (PEEK(offset& + 95) AND 7)
               POKE (offset& + 96), k% + (PEEK(offset& + 96) AND 7)
            ELSEIF (k% > 0) THEN
               POKE (offset& - 96), k%
               POKE (offset& - 95), k%
               POKE (offset& - 94), k%
               POKE (offset& - 1), k%
               POKE offset&, k%
               POKE (offset& + 1), k%
               POKE (offset& + 94), k%
               POKE (offset& + 95), k%
               POKE (offset& + 96), k%
            END IF

            offset& = (offset& - (95 * 3))

         NEXT z%

      NEXT x%
   NEXT y%

   ' ---------------------------
   ' Rotate the perspective view
   ' ---------------------------

   perang% = (perang% + 1)
   IF (perang% > 359) THEN perang% = 0

   ' ----------------------------------
   ' Paste buffer onto perspective view
   ' ----------------------------------

   PUT (175, 1), perbuff%, PSET

LOOP

REM $STATIC
SUB ButtonWait

' ---------------------------------------
' Wait until the mouse button is released
' ---------------------------------------

DO

   ax% = 3
   MouseDriver ax%, bx%, cx%, dx%
   lb% = ((bx% AND 1) <> 0)

LOOP WHILE (lb% = -1)

END SUB

SUB Drawegrid

' ------------------------
' Make the mouse invisible
' ------------------------

ax% = 2
MouseDriver ax%, 0, 0, 0

' -----------------------
' Redraw the editing grid
' -----------------------

FOR i% = 0 TO 15
   FOR j% = 0 TO 15

      x1% = (175 + (i% * 6))
      y1% = (103 + (j% * 6))

      ' -------------------------------------
      ' Fill in the block in the editing grid
      ' -------------------------------------

      SELECT CASE grid%
         CASE 0: c% = voxel%(gridx%, i%, (15 - j%))
         CASE 1: c% = voxel%(i%, (15 - gridy%), (15 - j%))
         CASE 2: c% = voxel%((15 - j%), i%, gridz%)
      END SELECT

      LINE (x1%, y1%)-(x1% + 4, y1% + 4), c%, BF
      
      IF c% = 0 THEN PSET (x1% + 2, y1% + 2), 15

   NEXT j%
NEXT i%

' ----------------------
' Make the mouse visible
' ----------------------

ax% = 1
MouseDriver ax%, 0, 0, 0

END SUB

SUB Drawgridbuts

   ' ------------------------
   ' Make the mouse invisible
   ' ------------------------

   ax% = 2
   MouseDriver ax%, 0, 0, 0

   LOCATE 18, 16: COLOR 2: PRINT "X"
   LOCATE 20, 16: COLOR 4: PRINT "Y"
   LOCATE 22, 16: COLOR 1: PRINT "Z"

   FOR i% = 0 TO 2
      LINE (115, 134 + (i% * 16))-(131, 144 + (i% * 16)), 8, B
   NEXT

   LINE (115, 134 + (grid% * 16))-(131, 144 + (grid% * 16)), 15, B
  
   COLOR 15

   SELECT CASE grid%
      CASE 0: LOCATE 23, 18: PRINT gridx%
      CASE 1: LOCATE 23, 18: PRINT gridy%
      CASE 2: LOCATE 23, 18: PRINT gridz%
   END SELECT
  
   ' ----------------------
   ' Make the mouse visible
   ' ----------------------

   ax% = 1
   MouseDriver ax%, 0, 0, 0

END SUB

SUB Drawiview

' -------------------------
' Redraw the isometric view
' -------------------------

LINE (38, 23)-(132, 132), 0, BF

DEF SEG = &HA000

   FOR z% = 0 TO 15
      FOR y% = 0 TO 15
         FOR x% = 0 TO 15

            k% = voxel%(x%, y%, z%)
           
            x1% = 40 + ((x% * 3) + (y% * 3))
            y1% = 100 + ((y% * 2) - (x% * 2) - (z% * 3))
            offset& = (y1% * 320&) + x1%

            IF (k% = 56) THEN
               k% = 56
               POKE (offset& - 321), k% + (PEEK(offset& - 321) AND 7)
               POKE (offset& - 320), k% + (PEEK(offset& - 320) AND 7)
               POKE (offset& - 319), k% + (PEEK(offset& - 319) AND 7)
               POKE (offset& - 1), k% + (PEEK(offset& - 1) AND 7)
               POKE offset&, k% + (PEEK(offset&) AND 7)
               POKE (offset& + 1), k% + (PEEK(offset& + 1) AND 7)
               POKE (offset& + 319), k% + (PEEK(offset& + 319) AND 7)
               POKE (offset& + 320), k% + (PEEK(offset& + 320) AND 7)
               POKE (offset& + 321), k% + (PEEK(offset& + 321) AND 7)
            ELSEIF (k% > 0) THEN
               POKE (offset& - 321), k%
               POKE (offset& - 320), k%
               POKE (offset& - 319), k%
               POKE (offset& - 1), k%
               POKE offset&, k%
               POKE (offset& + 1), k%
               POKE (offset& + 319), k%
               POKE (offset& + 320), k%
               POKE (offset& + 321), k%
            END IF

         NEXT x%
      NEXT y%
   NEXT z%

DEF SEG

' ----------
' Draw zgrid
' ----------

y1% = 100 - (gridz% * 3)

LINE (40, y1%)-(85, y1% + 30), 1
LINE (85, y1% + 30)-(130, y1%), 1
LINE (40, y1%)-(85, y1% - 30), 1
LINE (85, y1% - 30)-(130, y1%), 1

' ----------
' Draw xgrid
' ----------

x1% = 40 + (gridx% * 3)
y1% = 100 - (gridx% * 2)

LINE (x1%, y1% - 45)-(x1%, y1%), 2
LINE (x1%, y1%)-(x1% + 45, y1% + 30), 2
LINE (x1% + 45, y1% - 15)-(x1% + 45, y1% + 30), 2
LINE (x1%, y1% - 45)-(x1% + 45, y1% - 15), 2

' ----------
' Draw ygrid
' ----------

x1% = 85 - (gridy% * 3)
y1% = 130 - (gridy% * 2)

LINE (x1%, y1% - 45)-(x1%, y1%), 4
LINE (x1%, y1%)-(x1% + 45, y1% - 30), 4
LINE (x1% + 45, y1% - 75)-(x1% + 45, y1% - 30), 4
LINE (x1%, y1% - 45)-(x1% + 45, y1% - 75), 4


END SUB

DEFLNG A-Z
SUB MouseDriver (ax%, bx%, cx%, dx%)
 
   ' ------------
   ' Mouse driver
   ' ------------

   DEF SEG = VARSEG(mouse$)
   mouse% = SADD(mouse$)
   CALL Absolute(ax%, bx%, cx%, dx%, mouse%)

END SUB

